unit ENGINES; {Unit 6 of SEASUAD.PAS}

interface

  uses
    GLOBALS, MATHS;

  procedure Convolution (var alpha, beta: seasVector;
                  p, k: integer);

  procedure Laurent (var mu, gamma: seasVector;
                  q: integer);

  procedure MakeSigmaVec (var sigmaVec: seasVector;
                  var s: integer);

  procedure MakeRhoVec (var rhoVec: seasVector;
                  var rho: real;
                  var s: integer);

  procedure CompoundQuads (var sigmaVec, rhoVec: seasVector;
                  var rho, zeta: real;
                  var sPlus, quadNo: integer);

  procedure NewCompoundQuads (var sigmaVec, rhoVec: seasVector;
                  var zeta: vector;
                  var rho: real;
                  var quadNo: integer);

  procedure WKCoefficients (var phi, theta: seasVector;
                  var s: integer;
                  var lambda: real);

  function SeasResponse (var phi, theta: seasVector;
                  var s: integer;
                  var omega: real): real;

  function DualSeaResponse (var phi1, phi2, theta1, theta2: seasVector;
                  var s: integer;
                  var omega: real): real;

  function TripleSeaResponse (var phi0, phi1, phi2, theta0, theta1, theta2: seasVector;
                  var s: integer;
                  var omega: real): real;

  procedure GetFilterGainVector (var SCREENBOX: IBOX;
                  var dataBOX: rBOX;
                  var dataVector: longVector;
                  var phi, theta, phi1, theta1, phi2, theta2: seasVector;
                  var sPlus: integer);

implementation

{_________________________________________________}

  procedure Convolution (var alpha, beta: seasVector;
                  p, k: integer);
    var
      gamma: extended;
      i, j, r, s: integer;

  begin

    for j := p + k downto 0 do
      begin {j}
        s := Min(j, p);
        r := Max(0, j - k);
        gamma := 0.0;
        for i := r to s do
          gamma := gamma + alpha[i] * beta[j - i];
        beta[j] := gamma;
      end; {j}

  end; {Convolution}

{_________________________________________________}

  procedure Laurent (var mu, gamma: seasVector;
                  q: integer);

    var
      i, j: integer;

  begin
    for i := 0 to q do
      begin {i}
        gamma[i] := 0.0;
        for j := 0 to q - i do
          gamma[i] := gamma[i] + mu[j] * mu[j + i];
      end; {i}
  end; {Laurent}

 {_________________________________________________}

  procedure MakeSigmaVec (var sigmaVec: seasVector;
                  var s: integer);
{EA}
    var
      i: integer;
  begin
    for i := 0 to s - 1 do
      sigmaVec[i] := 1.0;
  end; {MakeSigmaVec}

{_________________________________________________}

  procedure MakeRhoVec (var rhoVec: seasVector;
                  var rho: real;
                  var s: integer);
    var
      i: integer;
      rhoPower: real;

  begin
    rhoPower := 1.0;
    for i := 0 to s - 1 do
      begin
        rhoVec[i] := rhoPower;
        rhoPower := rhoPower * rho;
      end;
  end; {MakeRhoVec}

{_________________________________________________}

  procedure CompoundQuads (var sigmaVec, rhoVec: seasVector;
                  var rho, zeta: real;
                  var sPlus, quadNo: integer);
    var
      quadZeros, quadPoles: SeasMatrix;
      omega: seasVector;
      degree: extended;
      alpha: seasVector;
      j, p, k: integer;

  begin

    degree := pi / 180;
    omega[1] := pi / 6.0;
    omega[2] := pi / 3.0;
    omega[3] := pi / 2.0;
    omega[4] := 2.0 * pi / 3.0;
    omega[5] := 5.0 * pi / 6.0;
    zeta := zeta * degree;

    QuadZeros[0, 0] := 1.0;
    QuadZeros[0, 1] := -1.0;
    QuadZeros[0, 2] := 0.0;

    for j := 1 to 5 do
      begin
        QuadZeros[j, 0] := 1.0;
        QuadZeros[j, 1] := -2.0 * Cos(omega[j] + zeta);
        QuadZeros[j, 2] := 1.0;
      end;

    QuadZeros[6, 0] := 1.0;
    if twoPass then
      begin
        QuadZeros[6, 1] := -2.0 * Cos(pi + zeta);
        QuadZeros[6, 2] := 1.0;
      end;
    if ordinary then
      begin
        QuadZeros[6, 1] := 1.0;
        QuadZeros[6, 2] := 0.0;
      end;

    sigmaVec[0] := QuadZeros[1, 0];
    sigmaVec[1] := QuadZeros[1, 1];
    sigmaVec[2] := QuadZeros[1, 2];

    p := 2;
    k := 2; {or is it unity?}
    for j := 2 to quadNo do
      begin
        alpha[0] := quadZeros[j, 0];
        alpha[1] := quadZeros[j, 1];
        alpha[2] := quadZeros[j, 2];
        Convolution(alpha, sigmaVec, p, k);
        k := k + 2
      end;

{++++++++++++++++++++++++++++++++++++++}

    quadPoles[0, 0] := 1.0;
    quadPoles[0, 1] := -rho;
    quadPoles[0, 2] := 0.0;

    for j := 1 to 5 do
      begin
        quadPoles[j, 0] := 1.0;
        quadPoles[j, 1] := -2.0 * rho * Cos(omega[j] + zeta);
        quadPoles[j, 2] := rho * rho;
      end;

    quadPoles[6, 0] := 1.0;
    if twoPass then
      begin
        quadPoles[6, 1] := -2.0 * rho * Cos(pi + zeta);
        quadPoles[6, 2] := rho * rho;
      end;
    if ordinary then
      begin
        quadPoles[6, 1] := rho;
        quadPoles[6, 2] := 0.0;
      end;

    rhoVec[0] := quadPoles[1, 0];
    rhoVec[1] := quadPoles[1, 1];
    rhoVec[2] := quadPoles[1, 2];

    p := 2;
    k := 2;
    for j := 2 to quadNo do
      begin
        alpha[0] := quadPoles[j, 0];
        alpha[1] := quadPoles[j, 1];
        alpha[2] := quadPoles[j, 2];
        Convolution(alpha, rhoVec, p, k);
        k := k + 2
      end;

{for j := 0 to 12 do}
{Writeln('    rhoVec[', j : 2, '] = ', rhoVec[j] : 6 : 4);}
{READLN;}

    noModel := false;

  end; {CompoundQuads}

{_________________________________________________}

  procedure NewCompoundQuads (var sigmaVec, rhoVec: seasVector;
                  var zeta: vector;
                  var rho: real;
                  var quadNo: integer);

    var
      quadZeros, quadPoles: SeasMatrix;
      omega: seasVector;
      degree: extended;
      zetaR: reaL;
      alpha: seasVector;
      j, p, k: integer;

  begin

    degree := pi / 180;
    omega[1] := pi / 6.0;
    omega[2] := pi / 3.0;
    omega[3] := pi / 2.0;
    omega[4] := 2.0 * pi / 3.0;
    omega[5] := 5.0 * pi / 6.0;

    QuadZeros[0, 0] := 1.0;
    QuadZeros[0, 1] := -1.0;
    QuadZeros[0, 2] := 0.0;

    for j := 1 to 5 do
      begin
        zetaR := zeta[j] * degree;
        QuadZeros[j, 0] := 1.0;
        QuadZeros[j, 1] := -2.0 * Cos(omega[j] + zetaR);
        QuadZeros[j, 2] := 1.0;
      end;

    QuadZeros[6, 0] := 1.0;
    if offset = true then
      begin
        zetaR := zeta[6] * degree;
        QuadZeros[6, 1] := -2.0 * Cos(pi + zetaR);
        QuadZeros[6, 2] := 1.0;
      end
    else
      begin
        QuadZeros[6, 1] := 1.0;
        QuadZeros[6, 2] := 0.0;
      end;

    sigmaVec[0] := QuadZeros[1, 0];
    sigmaVec[1] := QuadZeros[1, 1];
    sigmaVec[2] := QuadZeros[1, 2];

    p := 2;
    k := 2;
    for j := 2 to quadNo do
      begin
        alpha[0] := quadZeros[j, 0];
        alpha[1] := quadZeros[j, 1];
        alpha[2] := quadZeros[j, 2];
        Convolution(alpha, sigmaVec, p, k);
        k := k + 2
      end;

{++++++++++++++++++++++++++++++++++++++}

    quadPoles[0, 0] := 1.0;
    quadPoles[0, 1] := -rho;
    quadPoles[0, 2] := 0.0;

    for j := 1 to 5 do
      begin
        zetaR := zeta[j] * degree;
        quadPoles[j, 0] := 1.0;
        quadPoles[j, 1] := -2.0 * rho * Cos(omega[j] + zetaR);
        quadPoles[j, 2] := rho * rho;
      end;

    quadPoles[6, 0] := 1.0;
    if offset then
      begin
        zetaR := zeta[6] * degree;
        quadPoles[6, 1] := -2.0 * rho * Cos(pi + zetaR);
        quadPoles[6, 2] := rho * rho;
      end
    else
      begin
        quadPoles[6, 1] := rho;
        quadPoles[6, 2] := 0.0;
      end;

    rhoVec[0] := quadPoles[1, 0];
    rhoVec[1] := quadPoles[1, 1];
    rhoVec[2] := quadPoles[1, 2];

    p := 2;
    k := 2;
    for j := 2 to quadNo do
      begin
        alpha[0] := quadPoles[j, 0];
        alpha[1] := quadPoles[j, 1];
        alpha[2] := quadPoles[j, 2];
        Convolution(alpha, rhoVec, p, k);
        k := k + 2
      end;

{for j := 0 to 12 do}
{Writeln('    rhoVec[', j : 2, '] = ', rhoVec[j] : 6 : 4);}
{READLN;}

    noModel := false;

  end; {NewCompoundQuads}

{_________________________________________________}

  procedure WKCoefficients (var phi, theta: seasVector;
                  var s: integer;
                  var lambda: real);
    var
      j: integer;
      thetaSum, phiSum, factor: extended;

  begin

{Make lambda* R(z)R(1/z) + S(z)S(1/z)}
    for j := 0 to s do
      phi[j] := lambda * phi[j] + theta[j];

    thetaSum := theta[0];
    phiSum := phi[0];
    for j := 1 to s do
      begin
        thetaSum := thetaSum + 2 * theta[j];
        phiSum := phiSum + 2 * phi[j];
      end;

    factor := phiSum / thetaSum;
    for j := 0 to s do
      theta[j] := theta[j] * factor;

  end; {WKCoefficients}

{_________________________________________________}

  function SeasResponse (var phi, theta: seasVector;
                  var s: integer;
                  var omega: real): real;
    var
      numerator, denominator: extended;
      j: integer;

  begin
    numerator := theta[0];
    denominator := phi[0];
    for j := 1 to s do
      begin
        numerator := numerator + 2 * theta[j] * cos(j * omega);
        denominator := denominator + 2 * phi[j] * cos(j * omega);
      end;
    SeasResponse := numerator / denominator;

  end; {SeasResponse}

{_________________________________________________}

  function DualSeaResponse (var phi1, phi2, theta1, theta2: seasVector;
                  var s: integer;
                  var omega: real): real;
    var
      numerator, denominator, firstResponse: extended;
      j: integer;

  begin

    firstResponse := 1.0;

    numerator := theta1[0];
    denominator := phi1[0];
    for j := 1 to 10 do
      begin
        numerator := numerator + 2 * theta1[j] * cos(j * omega);
        denominator := denominator + 2 * phi1[j] * cos(j * omega);
      end;
    firstResponse := numerator / denominator;

    numerator := theta2[0];
    denominator := phi2[0];
    for j := 1 to 12 do
      begin
        numerator := numerator + 2 * theta2[j] * cos(j * omega);
        denominator := denominator + 2 * phi2[j] * cos(j * omega);
      end;
    DualSeaResponse := firstResponse * numerator / denominator;

  end; {DualSeaResponse}

{_________________________________________________}

  function TripleSeaResponse (var phi0, phi1, phi2, theta0, theta1, theta2: seasVector;
                  var s: integer;
                  var omega: real): real;
    var
      numerator, denominator, firstResponse: extended;
      j: integer;
{label}
{999;}

  begin

    firstResponse := 1.0;

    numerator := theta1[0];
    denominator := phi1[0];
    for j := 1 to 10 do
      begin
        numerator := numerator + 2 * theta1[j] * cos(j * omega);
        denominator := denominator + 2 * phi1[j] * cos(j * omega);
      end;
    firstResponse := numerator / denominator;


    numerator := theta2[0];
    denominator := phi2[0];
    for j := 1 to 12 do
      begin
        numerator := numerator + 2 * theta2[j] * cos(j * omega);
        denominator := denominator + 2 * phi2[j] * cos(j * omega);
      end;
    firstResponse := firstResponse * numerator / denominator;

    numerator := theta0[0];
    denominator := phi0[0];
    for j := 1 to 12 do
      begin
        numerator := numerator + 2 * theta0[j] * cos(j * omega);
        denominator := denominator + 2 * phi0[j] * cos(j * omega);
      end;
    TripleSeaResponse := firstResponse * numerator / denominator;


  end; {TripleSeaResponse}

{_________________________________________________}

  procedure GetFilterGainVector (var SCREENBOX: IBOX;
                  var dataBOX: rBOX;
                  var dataVector: longVector;
                  var phi, theta, phi1, theta1, phi2, theta2: seasVector;
                  var sPlus: integer);
    var
      BOXWDT, i: integer;
      yMax, yMin, omega, pi, inc: real;

  begin
    pi := 4.0 * Arctan(1.0);
    BOXWDT := SCREENBOX.X2 - SCREENBOX.X1;

    inc := pi / BOXWDT;
    omega := 0.0;

    yMax := 0.0;
    yMin := 0.0;
    for i := 0 to BOXWDT do
      begin {i}
        if ordinary then
          dataVector[i] := SeasResponse(phi, theta, sPlus, omega);
        if twoPass then
          dataVector[i] := DualSeaResponse(phi1, phi2, theta1, theta2, sPlus, omega);
        if threePass = true then
          dataVector[i] := TripleSeaResponse(phi, phi1, phi2, theta, theta1, theta2, sPlus, omega);

        yMax := RMax(yMax, dataVector[i]);
        yMin := RMin(yMin, dataVector[i]);
        omega := omega + inc;

{Writeln('    dataVector[', i : 2, ']: = ', dataVector[i] : 8 : 4);}
      end; {i}

    dataBOX.NOPUNCT := BOXWDT;
{dataBOX.xRange := BOXWDT;}
    dataBOX.xMax := pi;
    dataBOX.xMin := 0.0;
    dataBOX.yMax := yMax;
    dataBOX.yMin := yMin;
    dataBOX.xRB := pi;
    dataBOX.xLB := 0.0;
    dataBOX.yUB := yMax;
    dataBOX.yLB := yMin;

  end; {GetFilterGainVector}

 {________________________________________________}

end. {ENGINES: Unit 6 of SEASUAD.PAS}

